#include "DCMotor.h"

DCMotor::DCMotor(uint8_t _id, uint8_t _gear, bool _reverse, uint8_t _inertFactor) {
    this->motorID = _id;
    this->gear = _gear;
    this->reverse = _reverse;
    this->inertFactor = _inertFactor;
}

void DCMotor::attachEncoder(BaseTickEncoder* _BaseTickEncoder) {
    if (this->encoder)
        return;
    this->encoder = _BaseTickEncoder;
}

void DCMotor::attachEncoder(OpticalTickEncoder* _OpticalTickEncoder) {
    if (this->encoder)
        return;
    this->encoder = _OpticalTickEncoder;
}

void DCMotor::attachEncoder(HallTickEncoder* _HallTickEncoder) {
    if (this->encoder)
        return;
    this->encoder = _HallTickEncoder;
}

uint8_t DCMotor::getSpeed() {
    if (this->forwardSpeed > 0) {
        return this->forwardSpeed;
    } else if(this->backwardSpeed > 0) {
        return this->backwardSpeed;
    }

    return 0;
}

bool DCMotor::isStopped() {
    return this->state == STATE_STOPPED;
}

bool DCMotor::isForward() {
    return this->state == STATE_FORWARD;
}

bool DCMotor::isBackward() {
    return this->state == STATE_BACKWARD;
}

bool DCMotor::isBusy() {
    if (
            (      this->encoder
                && this->encoder->isLimited()
                && !this->encoder->isLimitExceeded()
            )
            || this->task != MOTOR_TASK_IDLE
    ) {
        return true;
    } 

    return false;
}

void DCMotor::forward(uint8_t speed) {
    if (this->isBusy())
        return;

    this->encoderEnabled = false;
    this->setForwardSpeed(speed);
}

void DCMotor::forward(uint8_t speed, float rotations) {
    if (this->isBusy() || (rotations <= 0) || (!this->encoder))
        return;

    this->encoderEnabled = true;
    this->encoder->setLimit(rotations * this->gear, true);
    this->setForwardSpeed(speed);
}

void DCMotor::backward(uint8_t speed) {
    if (this->isBusy())
        return;

    this->encoderEnabled = false;
    this->setBackwardSpeed(speed);
}

void DCMotor::backward(uint8_t speed, float rotations) {
    if (this->isBusy() || (rotations <= 0) || (!this->encoder))
        return;

    this->encoderEnabled = true;
    this->encoder->setLimit(rotations * this->gear, false);
    this->setBackwardSpeed(speed);
}

void DCMotor::stop(bool hard) {
    if (hard && (this->forwardSpeed > 0 || this->backwardSpeed > 0)) {
        uint16_t stopTime = this->getHardStopTime();

        Serial.println(stopTime);

        if (this->forwardSpeed > 0) {
            this->state = STATE_FORWARD;
            this->backward(100);
            this->delayAndOff(stopTime);

        } else if (this->backwardSpeed > 0) {
            this->state = STATE_BACKWARD;
            this->forward(100);
            this->delayAndOff(stopTime);
        }
    } else {
        this->off();
    }
}

void DCMotor::externalTick(uint16_t periodMs) {
    if (this->task != MOTOR_TASK_IDLE) {
        this->msCounter += periodMs;

        if (this->msCounter >= this->msLimit) {
            this->msCounter = 0;
            this->msLimit = 0;

            if (this->task == MOTOR_TASK_OFF) {
                this->off();
            }
            this->task = MOTOR_TASK_IDLE;
        }
    }

    if ((this->encoder)
        && this->state != STATE_STOPPED
    ) {
        this->checkEncoderLimit();
    }
}

void DCMotor::abort() {
    if (this->encoder) {
        this->encoder->disable();
        this->encoder->reset();
    }

    this->state = STATE_STOPPED;
    this->task = MOTOR_TASK_IDLE;
}

uint8_t DCMotor::getGear() {
    return this->gear;
}

/*****************************************************************/
/***********************     Private     *************************/
/*****************************************************************/

void DCMotor::setForwardSpeed(uint8_t speed) {
    speed = this->normalizeSpeed(speed);
    this->state = STATE_FORWARD;
    this->backwardSpeed = 0;
    if (this->forwardSpeed == speed) 
        return;
    this->forwardSpeed = speed;
    bool motor_reverse = this->reverse ? true : false;
    this->writeSpeed(speed, motor_reverse);
}

void DCMotor::setBackwardSpeed(uint8_t speed) {
    speed = this->normalizeSpeed(speed);
    this->state = STATE_BACKWARD;
    this->forwardSpeed = 0;
    if (this->backwardSpeed == speed) { return; }
    this->backwardSpeed = speed;
    bool motor_reverse = this->reverse ? false : true;
    this->writeSpeed(speed, motor_reverse);
}

void DCMotor::checkEncoderLimit() {
    if (this->encoder) {
        if (this->isEncoderLimitExceeded()) {
            this->stop(true);
            this->encoder->disable();
            this->encoder->reset();
        }
    }
}

bool DCMotor::isEncoderLimitExceeded() {
    if (this->encoder) {
        return this->encoderEnabled
            && this->encoder
            && this->encoder->isLimited()
            && this->encoder->isLimitExceeded();
    }
}

void DCMotor::off() {
    this->state = STATE_STOPPED;
    this->forwardSpeed = 0;
    this->backwardSpeed = 0;

    this->writeSpeed(0, this->reverse);

    if (this->encoder) {
        this->encoder->disable();
    }
}

void DCMotor::delayAndOff(uint8_t time_ms) {
    this->msLimit = (uint32_t)time_ms;
    this->task = MOTOR_TASK_OFF;
}

uint8_t DCMotor::getHardStopTime() {
    uint8_t speed = max(this->backwardSpeed, this->forwardSpeed);
    return (uint8_t)(( (speed/2 + (100 - speed)) * inertFactor)/100);
}

uint8_t DCMotor::normalizeSpeed(uint8_t speed) {
    speed = constrain(speed, 0, 100);
    return speed;
}

void DCMotor::writeSpeed(uint8_t speed, bool _reverse) {
    speed = constrain(speed, 0, 100);
    int16_t pwmValue = (int16_t)(((int16_t)speed * 255)/100);
    Board::external()->writeDCMotor(this->motorID, pwmValue, _reverse);
}