#include "DualWheelSpeedRegulator.h"

//#define DEBUG_MODE

DualWheelSpeedRegulator::DualWheelSpeedRegulator() {
	this->pidRegulator = new PidRegulator();
	this->pidRegulator->setResultLimits(MIN_CORRECTION_LIMIT, MAX_CORRECTION_LIMIT);
}

void DualWheelSpeedRegulator::setDebugMode(char mode) {
	this->pidRegulator->setDebugMode(mode);
}

void DualWheelSpeedRegulator::attachLeftMotor(DCMotor* motor) {
	this->leftMotor = motor;
}

void DualWheelSpeedRegulator::attachRightMotor(DCMotor* motor) {
	this->rightMotor = motor;
}

void DualWheelSpeedRegulator::attachLeftEncoder(OpticalTickEncoder* encoder) {
	this->leftEncoder = encoder;
}

void DualWheelSpeedRegulator::attachRightEncoder(OpticalTickEncoder* encoder) {
	this->rightEncoder = encoder;
}

void DualWheelSpeedRegulator::attachLeftEncoder(HallTickEncoder* encoder) {
	this->leftEncoder = encoder;
}

void DualWheelSpeedRegulator::attachRightEncoder(HallTickEncoder* encoder) {
	this->rightEncoder = encoder;
}

void DualWheelSpeedRegulator::attachLeftEncoder(BaseTickEncoder* encoder) {
	this->leftEncoder = encoder;
}

void DualWheelSpeedRegulator::attachRightEncoder(BaseTickEncoder* encoder) {
	this->rightEncoder = encoder;
}

void DualWheelSpeedRegulator::reset() {
	this->pidRegulator->reset();
}

void DualWheelSpeedRegulator::setSpeed(int8_t speed) {
	this->speed = speed;
}

void DualWheelSpeedRegulator::setPidFactors(float Kp, float Ki, float Kd) {
	this->pidRegulator->setPidFactors(Kp, Ki, Kd);
}

void DualWheelSpeedRegulator::throttlingOn() {
    this->throttlingMode = true;
}

void DualWheelSpeedRegulator::throttlingOff() {
    this->throttlingMode = false;
}

void DualWheelSpeedRegulator::run() {
	#ifdef DEBUG_MODE
		Serial.println("\n2WD_SPEED_REGULTOR [");
	#endif

	// Check if one or both encoders are exceeded, than return
	// because it is too late and we can't regulate anything
	if (this->leftEncoder->isLimited() == false
		|| this->leftEncoder->isLimitExceeded()
		|| this->rightEncoder->isLimited() == false
		|| this->rightEncoder->isLimitExceeded()
	) {
		#ifdef DEBUG_MODE
			Serial.print("exit 1\n]");
		#endif

		return;
	}

	// Check if still no data from encoders,
	// than need return and try again later
	if (this->leftEncoder->getPulses() == 0
		&& this->rightEncoder->getPulses() == 0
	) {
		#ifdef DEBUG_MODE
			Serial.print("exit 2\n]");
		#endif

		return;
	}

	uint16_t leftSpeed = 0;
	uint16_t rightSpeed = 0;

	int16_t correction = this->pidRegulator->run(
			(int16_t)leftEncoder->getPulses(),
			(int16_t)rightEncoder->getPulses()
		);

	if (correction > 0) {
		// Left faster
		int16_t leftRange = speed;
		int16_t rightRange = 100 - speed;

		leftSpeed = speed - (leftRange * correction/(leftRange + rightRange));
		rightSpeed = speed + (rightRange * correction/(leftRange + rightRange));
	} else {
		// Right faster
		uint16_t leftRange = 100 - speed;
		uint16_t rightRange = speed;

		leftSpeed = speed + (leftRange * abs(correction)/(leftRange + rightRange));
		rightSpeed = speed - (rightRange * abs(correction)/(leftRange + rightRange));
	}

	leftSpeed = constrain(leftSpeed, 0, 100);
	rightSpeed = constrain(rightSpeed, 0, 100);

	#ifdef DEBUG_MODE
		Serial.print("\tCorr:"); Serial.print(correction);
		Serial.print(", New left="); Serial.print(leftSpeed);
		Serial.print("right="); Serial.print(rightSpeed); Serial.println("\n]");
	#endif

	if (speed < 0) {
		if (leftMotor->isStopped() == false) {
			leftMotor->setBackwardSpeed(leftSpeed);
		}

		if (rightMotor->isStopped() == false) {
			rightMotor->setBackwardSpeed(rightSpeed);
		}
	} else {
		if (leftMotor->isStopped() == false) {
			leftMotor->setForwardSpeed(leftSpeed);
		}

		if (rightMotor->isStopped() == false) {
			rightMotor->setForwardSpeed(rightSpeed);
		}
	}
}
