#include "DualWheelTruck.h"

DualWheelTruck* DualWheelTruck::instance = nullptr;

DualWheelTruck::DualWheelTruck(uint8_t wheelRadius_mm, uint16_t wheelBase_mm) {
    this->wheelRadius = wheelRadius_mm;
    this->wheelBase = wheelBase_mm;
    DualWheelTruck::instance = this;
}

void DualWheelTruck::asyncRun() {
    if (this->leftMotor) {
        this->leftMotor->externalTick(this->timerDelayMS);
    }

    if (this->rightMotor) {
        this->rightMotor->externalTick(this->timerDelayMS);
    }

    this->speedRegulatorCallTimer++;

    if (this->speedRegulatorCallTimer >= this->speedRegulatorCallPeriod) {
        this->speedRegulatorCallTimer = 0;
        this->correctSpeed();
    }
}

void DualWheelTruck::init() {
    if (!this->speedRegulator) {
        this->speedRegulator = new DualWheelSpeedRegulator();
    }

    if (!this->speedDiffRegulator) {
        this->speedDiffRegulator = new DualWheelSpeedDifferenceRegulator();
    }

    // Init interrupts for motors's and PID regulator's and etc.
    Tweak::attachMsCallback(timerHandler, (uint16_t)this->timerDelayMS);
}

void DualWheelTruck::timerHandler(void) {
    DualWheelTruck::me()->asyncRun();
}

void DualWheelTruck::setPidAlign(float Kp, float Ki, float Kd) {
    if (this->speedRegulator) {
        this->speedRegulator->setPidFactors(Kp, Ki, Kd);
    }
}

void DualWheelTruck::setPidDiff(float Kp, float Ki, float Kd) {
    if (this->speedDiffRegulator) {
        this->speedDiffRegulator->setPidFactors(Kp, Ki, Kd);
    }
}

void DualWheelTruck::setPidAlignDebug(char mode) {
    if (this->speedRegulator) {
        this->speedRegulator->setDebugMode(mode);
    }
}

void DualWheelTruck::setPidDiffDebug(char mode) {
    if (this->speedDiffRegulator) {
        this->speedDiffRegulator->setDebugMode(mode);
    }
}

void DualWheelTruck::attMotorL(uint8_t id, uint8_t gear, bool normalDirection, uint8_t inertFactor) {
    this->leftMotor = new DCMotor(id, gear, !normalDirection, inertFactor);

    if (this->speedRegulator) {
        this->speedRegulator->attachLeftMotor(this->leftMotor);
    }

    if (this->speedDiffRegulator) {
        this->speedDiffRegulator->attachLeftMotor(this->leftMotor);
    }
}

void DualWheelTruck::attMotorR(uint8_t id, uint8_t gear, bool normalDirection, uint8_t inertFactor) {
    this->rightMotor = new DCMotor(id, gear, !normalDirection, inertFactor);

    if (this->speedRegulator) {
        this->speedRegulator->attachRightMotor(this->rightMotor);
    }

    if (this->speedDiffRegulator) {
        this->speedDiffRegulator->attachRightMotor(this->rightMotor);
    }
}

void DualWheelTruck::attOptEncL(uint8_t ch, uint8_t sectorsPerLoop) {
    this->leftEncoder = new OpticalTickEncoder(ch, sectorsPerLoop);
    this->leftMotor->attachEncoder(this->leftEncoder);

    if (this->speedRegulator) {
        this->speedRegulator->attachLeftEncoder(this->leftEncoder);
    }

    if (this->speedDiffRegulator) {
        this->speedDiffRegulator->attachLeftEncoder(this->leftEncoder);
    }
}

void DualWheelTruck::attOptEncR(uint8_t ch, uint8_t sectorsPerLoop) {
    this->rightEncoder = new OpticalTickEncoder(ch, sectorsPerLoop);
    this->rightMotor->attachEncoder(this->rightEncoder);

    if (this->speedRegulator) {
        this->speedRegulator->attachRightEncoder(this->rightEncoder);
    }

    if (this->speedDiffRegulator) {
        this->speedDiffRegulator->attachRightEncoder(this->rightEncoder);
    }
}

void DualWheelTruck::attHallEncL(uint8_t ch1, uint8_t ch2, uint8_t sectorsPerLoop) {
    this->leftEncoder = new HallTickEncoder(ch1, ch2, sectorsPerLoop);

    this->leftEncoder->name = 2;
    pinMode(8, OUTPUT);
    digitalWrite(8, LOW);
    pinMode(A5, OUTPUT);
    digitalWrite(A5, LOW);

    this->leftMotor->attachEncoder(this->leftEncoder);

    if (this->speedRegulator) {
        this->speedRegulator->attachLeftEncoder(this->leftEncoder);
    }

    if (this->speedDiffRegulator) {
        this->speedDiffRegulator->attachLeftEncoder(this->leftEncoder);
    }
}

void DualWheelTruck::attHallEncR(uint8_t ch1, uint8_t ch2, uint8_t sectorsPerLoop) {
    this->rightEncoder = new HallTickEncoder(ch1, ch2, sectorsPerLoop);

    this->rightEncoder->name = 1;
    pinMode(4, OUTPUT);
    digitalWrite(4, LOW);
    pinMode(A4, OUTPUT);
    digitalWrite(A4, LOW);

    this->rightMotor->attachEncoder(this->rightEncoder);

    if (this->speedRegulator) {
        this->speedRegulator->attachRightEncoder(this->rightEncoder);
    }

    if (this->speedDiffRegulator) {
        this->speedDiffRegulator->attachRightEncoder(this->rightEncoder);
    }
}

void DualWheelTruck::enEncoders(){
    if (this->leftEncoder)
        this->leftEncoder->enable();
    if (this->rightEncoder)
        this->rightEncoder->enable();
}

void DualWheelTruck::disEncoders(){
    if (this->leftEncoder)
        this->leftEncoder->disable();
    if (this->rightEncoder)
        this->rightEncoder->disable();
}

uint32_t DualWheelTruck::getDistance(){
    if (this->leftEncoder && this->rightEncoder){
        uint32_t dist_left = this->leftEncoder->getDistanceTravelled(
                                                    this->wheelRadius, 
                                                    this->leftMotor->getGear()
                                                );
        uint32_t dist_right = this->rightEncoder->getDistanceTravelled(
                                                    this->wheelRadius, 
                                                    this->rightMotor->getGear()
                                                );
        return (dist_left + dist_right)/2;
    } else {
        #ifdef DEBUG_MODE
        // Serial.println(F(String(__func__ + ": encoders not set")));
        #endif /* DEBUG_MODE */
        return 0;
    }
}

void DualWheelTruck::resetDistance(){
    if (this->leftEncoder)
        this->leftEncoder->resetDistanceTravelled();
    if (this->rightEncoder)
        this->rightEncoder->resetDistanceTravelled();
}

// ****************************************************************
// ************************  LEFT MOTOR   *************************
// ****************************************************************

void DualWheelTruck::motorLFwd(uint8_t speed, uint16_t distance) {
    if (this->leftMotor) {
        if (distance > 0) {
            #ifdef DEBUG_MODE
                Serial.print("\nLeftMotorForward [\ndest="); Serial.print(distance);
                Serial.print("\nsectors=");
                Serial.print(
                    (uint16_t)this->leftEncoder->sectorsPerLoop
                        * this->distanceToWheelRotations(distance)
                );
                Serial.print("\nrot="); Serial.println(this->distanceToWheelRotations(distance));
                Serial.println("]");
            #endif

            this->leftMotor->forward(speed, this->distanceToWheelRotations(distance));
        } else {
            this->leftMotor->forward(speed);
        }

        this->disableSpeedCorrection();
    }
}

void DualWheelTruck::motorLBwd(uint8_t speed, uint16_t distance) {
    if (this->leftMotor) {
        if (distance > 0) {
            #ifdef DEBUG_MODE
                Serial.print("\nLeftMotorBackward [\ndest="); Serial.print(distance);
                Serial.print("\nsectors=");
                Serial.print(
                    (uint16_t)this->leftEncoder->sectorsPerLoop
                        * this->distanceToWheelRotations(distance)
                );
                Serial.print("\nrot="); Serial.println(this->distanceToWheelRotations(distance));
                Serial.println("]");
            #endif

            this->leftMotor->backward(speed, this->distanceToWheelRotations(distance));
        } else {
            this->leftMotor->backward(speed);
        }

        this->disableSpeedCorrection();
    }
}

void DualWheelTruck::motorLStop(bool hard) {
    if (this->leftMotor) {
        this->leftMotor->stop(hard);
        this->disableSpeedCorrection();
    }
}

// ****************************************************************
// ***********************  RIGHT MOTOR   *************************
// ****************************************************************

void DualWheelTruck::motorRFwd(uint8_t speed, uint16_t distance) {
    if (this->rightMotor) {
        if (distance > 0) {
            #ifdef DEBUG_MODE
                Serial.print("\nRghtMotorForward [\ndest="); Serial.print(distance);
                Serial.print("\nsectors=");
                Serial.print(
                    (uint16_t)this->rightEncoder->sectorsPerLoop
                       * this->distanceToWheelRotations(distance)
                );
                Serial.print("\nrot="); Serial.println(this->distanceToWheelRotations(distance));
                Serial.println("]");
            #endif

            this->rightMotor->forward(speed, this->distanceToWheelRotations(distance));
        } else {
            this->rightMotor->forward(speed);
        }

        this->disableSpeedCorrection();
    }
}

void DualWheelTruck::motorRBwd(uint8_t speed, uint16_t distance) {
    if (this->rightMotor) {
        if (distance > 0) {
            #ifdef DEBUG_MODE
                Serial.print("\nRightMotorBackward [\ndest="); Serial.print(distance);
                Serial.print("\nsectors=");
                Serial.print(
                    (uint16_t)this->rightEncoder->sectorsPerLoop
                        * this->distanceToWheelRotations(distance)
                );
                Serial.print("\nrot="); Serial.println(this->distanceToWheelRotations(distance));
                Serial.println("]");
            #endif

            this->rightMotor->backward(speed, this->distanceToWheelRotations(distance));
        } else {
            this->rightMotor->backward(speed);
        }

        this->disableSpeedCorrection();
    }
}

void DualWheelTruck::motorRStop(bool hard) {
    if (this->rightMotor) {
        this->rightMotor->stop(hard);
        this->disableSpeedCorrection();
    }
}

// ****************************************************************
// **********************  COMMON STUFF   *************************
// ****************************************************************

bool DualWheelTruck::isBusy() {
    return this->leftMotor->isBusy() || this->rightMotor->isBusy();
}

void DualWheelTruck::moveFwd(uint8_t speed, uint16_t distance) {
    this->state = MOVE_FORWARD;
    this->motorLFwd(speed, distance);
    this->motorRFwd(speed, distance);
    this->enableSpeedCorrection((int8_t)speed);
}

void DualWheelTruck::moveBwd(uint8_t speed, uint16_t distance) {
    this->state = MOVE_BACKWARD;
    this->motorLBwd(speed, distance);
    this->motorRBwd(speed, distance);
    this->enableSpeedCorrection(-1 * ((int8_t)speed) );
}

void DualWheelTruck::turnL(uint8_t speed, uint16_t radius, uint16_t angle) {
    this->state = TURN_LEFT;

    int8_t leftSpeed = 0;

    if (this->leftEncoder && this->rightEncoder) {
        // whillBase is in mm, so half of base in cm is: (this->wheelBase/10)/2 => this->wheelBase/20
        uint8_t halfBase = this->wheelBase/20;

        if (radius < halfBase) {
            // Left motor backward
            // Right motor forward
            leftSpeed = (int8_t)((-speed * (halfBase - radius))/(halfBase + radius));

            #ifdef DEBUG_MODE
                Serial.print("\nLeft turn(INNER): [\nleft wheel speed=");
                Serial.print(leftSpeed);
                Serial.print("\n left wheel distance=");
                Serial.print(this->turningAngleToDistance(halfBase - radius, angle));
            #endif

            this->motorLBwd(
                leftSpeed,
                this->turningAngleToDistance(halfBase - radius, angle)
            );
        } else if(radius == halfBase) {
            // Left motor stopped
            // Right motor forward
            #ifdef DEBUG_MODE
                Serial.print("\nLeft turn(ZERO): [\nleft wheel speed=0");
                Serial.print("\nleft wheel distance=0");
            #endif

            this->motorLStop(false);
        } else {
            // Left motor forward
            // Right motor forward
            leftSpeed = (int8_t)((speed * (radius - halfBase))/(radius + halfBase));

            #ifdef DEBUG_MODE
                Serial.print("\nLeft turn(OUTER): [\nleft wheel speed=0");
                Serial.print(leftSpeed);
                Serial.print("\nleft wheel distance=0");
                Serial.print(this->turningAngleToDistance(radius - halfBase, angle));
            #endif

            this->motorLFwd(
                leftSpeed,
                this->turningAngleToDistance(radius - halfBase, angle)
            );
        }

        #ifdef DEBUG_MODE
            Serial.print("\nright wheel speed=");
            Serial.print(speed);
            Serial.print("\nright wheel distance=0");
            Serial.println(this->turningAngleToDistance(radius + halfBase, angle));
            Serial.println("]");
        #endif

        this->motorRFwd(
            speed,
            this->turningAngleToDistance(radius + halfBase, angle)
        );

        if (leftSpeed != 0) {
            this->enableSpeedDifferenceCorrection(leftSpeed, speed);
        }
    }
}

void DualWheelTruck::turnR(uint8_t speed, uint16_t radius, uint16_t angle) {
    this->state = TURN_RIGHT;

    int8_t rightSpeed = 0;

    if (this->leftEncoder && this->rightEncoder) {
        // whillBase is in mm, so half of base in cm is: (this->wheelBase/10)/2 => this->wheelBase/20
        uint8_t halfBase = this->wheelBase/20;

        if (radius < halfBase) {
            // Left motor forward
            // Right motor backward
            rightSpeed = (int8_t)((-speed * (halfBase - radius))/(halfBase + radius));

            #ifdef DEBUG_MODE
                Serial.print("\nRight turn(INNER): [\nright wheel speed=");
                Serial.print(rightSpeed);
                Serial.print("\n right wheel distance=");
                Serial.print(this->turningAngleToDistance(halfBase - radius, angle));
            #endif

            this->motorRBwd(
                rightSpeed,
                this->turningAngleToDistance(halfBase - radius, angle)
            );
        } else if(radius == halfBase) {
            // Left motor forward
            // Right motor stopped
            #ifdef DEBUG_MODE
                Serial.print("\nRight turn(ZERO): [\nright wheel speed=0");
                Serial.print("\nright wheel distance=0");
            #endif

            this->motorRStop(false);
        } else {
            // Left motor forward
            // Right motor forward
            rightSpeed = (int8_t)((speed * (radius - halfBase))/(radius + halfBase));

            #ifdef DEBUG_MODE
                Serial.print("\nRight turn(OUTER): [\nright wheel speed=0");
                Serial.print(rightSpeed);
                Serial.print("\nright wheel distance=0");
                Serial.print(this->turningAngleToDistance(radius - halfBase, angle));
            #endif

            this->motorRFwd(
                rightSpeed,
                this->turningAngleToDistance(radius - halfBase, angle)
            );
        }

        #ifdef DEBUG_MODE
            Serial.print("\nleft wheel speed=");
            Serial.print(speed);
            Serial.print("\nleft wheel distance=0");
            Serial.println(this->turningAngleToDistance(radius + halfBase, angle));
            Serial.println("]");
        #endif

        this->motorLFwd(
            speed,
            this->turningAngleToDistance(radius + halfBase, angle)
        );

        if (rightSpeed != 0) {
            this->enableSpeedDifferenceCorrection(speed, rightSpeed);
        }
    }
}

// ****************************************************************
// **********************    private     **************************
// ****************************************************************

/**
 * @param uint16_t distance - cm (6553 cm max = 65m)
 * @param uint8_t wheelRadius - mm
 */
float DualWheelTruck::distanceToWheelRotations(uint16_t distance) {
//    return (float)(distance*10)/(TWO_PI * this->wheelRadius);
    return MotionMath::distanceToWheelRotations(distance, this->wheelRadius);
}

/**
 * Calculate how much wheel rotations need for wheel turn path
 * @param uint8_t turnRadius - cm
 * @param uint16_t angle - deg
 * @param uint8_t wheelRadius - mm
 */
float DualWheelTruck::turningAngleToWheelRotations(uint8_t radius, uint16_t angle) {
    // Wheel rotations Wr = L/C, where:
    // L - is wheel path length: L =  2 * Pi * turnRadius * angle/360
    // C - is wheel circle length: C = 2 * Pi * wheelRadius
    // Don't forget that whillRadius is mm scale, so we need devide it by 10.
    // So ttoal equation: Wr = (2 * Pi * turnRadius * angle/360) / (2 * Pi * wheelRadius/10)
    // But if we will reduce 2*Pi from equations 2 and 3, then result
    // equation will be much simpler: Wr = (turnRadius * angle/360) / (wheelRadius/10) => (turnRadius * angle/36) / wheelRadius
//    return (float)(radius * angle/36) / this->wheelRadius;
    return MotionMath::turningAngleToWheelRotations(radius, angle, this->wheelRadius);
}

/**
 * Calculate length of wheel path
 * @param uint8_t turnRadius - cm
 * @param uint16_t angle - deg
 */
float DualWheelTruck::turningAngleToDistance(uint8_t radius, uint16_t angle) {
//    return (float)(TWO_PI * radius * angle/360);
    return MotionMath::turningAngleToPathDistance(radius, angle);
}

void DualWheelTruck::disableSpeedCorrection() {
    this->speed = 0;
}

void DualWheelTruck::enableSpeedCorrection(int8_t speed) {
    if (this->leftEncoder
        && this->rightEncoder
        && speed != 0
    ) {

        if (this->speedDiffRegulator) {
            this->speedDiffRegulator->reset();
        }

        if (this->speedRegulator) {
            this->speedRegulator->reset();
            this->speedRegulator->setSpeed(speed);
        }

        this->speed = speed;
    }
}

void DualWheelTruck::enableSpeedDifferenceCorrection(int8_t leftSpeed, int8_t rightSpeed) {
    if (this->leftEncoder
        && this->rightEncoder
        && leftSpeed != 0
        && rightSpeed != 0
    ) {

        if (this->speedRegulator) {
            this->speedRegulator->reset();
        }

        if (this->speedDiffRegulator) {
            this->speedDiffRegulator->reset();
            this->speedDiffRegulator->setSpeedDifference(leftSpeed, rightSpeed);
        }
    }
}

bool DualWheelTruck::isSpeedCorrectionEnabled() {
    return this->speed != 0
        || this->state == TURN_LEFT
        || this->state == TURN_RIGHT;
}

void DualWheelTruck::correctSpeed() {
    if (this->isSpeedCorrectionEnabled() == false)
        return;

    if (this->state == MOVE_FORWARD || this->state == MOVE_BACKWARD) {
        if (this->speedRegulator) {
            this->speedRegulator->run();
        }
    }

    if (this->state == TURN_LEFT || this->state == TURN_RIGHT) {
        if (this->speedDiffRegulator) {
            this->speedDiffRegulator->run();
        }
    }
}
