#ifndef _EVERY_2WD_TRUCK_H
#define _EVERY_2WD_TRUCK_H

#include <Arduino.h>
#include <Tweak.h>
#include "DCMotor.h"
#include "PidRegulator.h"
#include "DualWheelSpeedRegulator.h"
#include "DualWheelSpeedDifferenceRegulator.h"
#include "MotionMath.h"

//#define DEBUG_MODE

enum TruckState {
    STOP,
    MOVE_FORWARD,
    MOVE_BACKWARD,
    TURN_LEFT,
    TURN_RIGHT,
};

class DualWheelTruck {
    public:
        DualWheelTruck(uint8_t wheelRadius_mm, uint16_t wheelBase_mm);

        static DualWheelTruck* instance;

        static DualWheelTruck * me() {
            return instance;
        }

        void init();

        void setPidAlign(float Kp, float Ki, float Kd);
        void setPidDiff(float Kp, float Ki, float Kd);
        void setPidAlignDebug(char mode);
        void setPidDiffDebug(char mode);

        void attMotorL(uint8_t id, uint8_t gear = 1, bool normalDirection = false, uint8_t inertFactor = 0);
        void attMotorR(uint8_t id, uint8_t gear = 1, bool normalDirection = true, uint8_t inertFactor = 0);

        void attOptEncL(uint8_t ch, uint8_t sectorsPerLoop);
        void attOptEncR(uint8_t ch, uint8_t sectorsPerLoop);
        void attHallEncL(uint8_t ch1, uint8_t ch2, uint8_t pulsesPerLoop);
        void attHallEncR(uint8_t ch1, uint8_t ch2, uint8_t pulsesPerLoop);

        void motorLFwd(uint8_t speed, uint16_t distance = 0);
        void motorLBwd(uint8_t speed, uint16_t distance = 0);
        void motorLStop(bool hard = true);

        void motorRFwd(uint8_t speed, uint16_t distance = 0);
        void motorRBwd(uint8_t speed, uint16_t distance = 0);
        void motorRStop(bool hard = true);

        void moveFwd(uint8_t speed, uint16_t distance = 0);
        void moveBwd(uint8_t speed, uint16_t distance = 0);
        void turnL(uint8_t speed, uint16_t radius, uint16_t angle);
        void turnR(uint8_t speed, uint16_t radius, uint16_t angle);
        bool isBusy();

        void enEncoders();
        void disEncoders();
        uint32_t getDistance();
        void resetDistance();

        static void timerHandler(void);

    private:

        TruckState state = STOP;

        // Motors
        DCMotor* leftMotor;
        DCMotor* rightMotor;

        // Encoders
        BaseTickEncoder* leftEncoder;
        BaseTickEncoder* rightEncoder;

        // 2 WD forward-backward speed regulator
        DualWheelSpeedRegulator* speedRegulator;
        // 2 WD turn left-right speed regulator
        DualWheelSpeedDifferenceRegulator* speedDiffRegulator;

        uint16_t leftEncoderCounter = 0;
        uint16_t rightEncoderCounter = 0;

        // Remember speed. Signed value from -100% to 100%
        int8_t speed = 0;
        uint8_t speedRegulatorCallPeriod = 20;
        uint8_t speedRegulatorCallTimer = 0;

        uint8_t timerDelayMS = 1;

        uint8_t wheelRadius = 0; // millimeters
        uint16_t wheelBase = 0; // millimeters

        void asyncRun();
        void correctSpeed();
        void disableSpeedCorrection();
        void enableSpeedCorrection(int8_t speed);
        void enableSpeedDifferenceCorrection(int8_t leftSpeed, int8_t rightSpeed);
        bool isSpeedCorrectionEnabled();
        float distanceToWheelRotations(uint16_t distance);
        float turningAngleToWheelRotations(uint8_t radius, uint16_t angle);
        float turningAngleToDistance(uint8_t radius, uint16_t angle);
};

#endif
