#include "SmoothServo.h"

/*******************************************************************************
 ***************************     Constructor    ********************************
 ******************************************************************************/
SmoothServo::SmoothServo(
    uint8_t pin,
    servoEnvironment mode,
    uint16_t min,
    uint16_t max,
    uint16_t minSpeed,
    uint16_t maxSpeed,
    uint16_t acceleration
) {
    this->env = mode;
    this->servoPin = pin;
    this->servo_min = min;
    this->servo_max = max;
    this->minAngularSpeed = minSpeed;
    this->maxAngularSpeed = maxSpeed;
    this->acceleration = acceleration;
    switch (this->env) {
        case BOARD:
            /* Board configured outside */
            this->write = &SmoothServo::write_board;
            break;
        case ARDUINO_SERVO:
            this->arduinoServoClient = new Servo();
            this->arduinoServoClient->attach(this->servoPin, this->servo_min, this->servo_max);
            this->write = &SmoothServo::write_arduino;
            break;
        case PCA9685:
            this->i2c_cli = new I2C();
            I2CTransport::setClient(this->i2c_cli);
            I2CTransport::begin();
            /* Singleton, needs init */
            PCA9685::begin();
            this->write = &SmoothServo::write_pca9685;
            break;
        case STM8_I2C:
            this->i2c_cli = new I2C();
            I2CTransport::setClient(this->i2c_cli);
            I2CTransport::begin();
            /* Singleton, doesn't need init */
            this->write = &SmoothServo::write_stm8_i2c;
            break;
        default:
            break;
    }
}

/*******************************************************************************
 ****************************     Software     *********************************
 ******************************************************************************/

void SmoothServo::begin(uint8_t ms, uint8_t startAngle) {
    this->internalTimerMs = ms;
    this->rotate(startAngle);
    if (ms > 0) {
        this->tweak_id = Tweak::attachMsMemberCallback(this, ms);
    }
}

void SmoothServo::setMinAngularSpeed(uint8_t minSpeed) {
    this->minAngularSpeed = minSpeed;
}

void SmoothServo::setMaxAngularSpeed(uint16_t maxSpeed) {
    this->maxAngularSpeed = maxSpeed;
}

void SmoothServo::setMaxAcceleration(uint16_t acceleration) {
    this->acceleration = acceleration;
}

void SmoothServo::rotate(int16_t angle) {
    angle = normalizeAngle(angle);
    this->destAngle = angle;
    setAngle(angle);
}

void SmoothServo::smoothRotate(int16_t angle) {
    angle = normalizeAngle(angle);

    if (this->currentAngle == -1 || this->currentAngle == angle) { return; }

    this->destAngle = angle;
    this->speedDownAngle = this->destAngle;
    this->middleAngle = (this->currentAngle + this->destAngle) / 2;
    this->currentAngularSpeedMsScale = (uint32_t)this->minAngularSpeed * MS_FACTOR;

    if (this->currentAngle > angle) {
        /* CW rotation */
        this->isCCW = false;
    } else {
        /* CCW rotation */
        this->isCCW = true;
    }
    this->isBusy_var = true;
}

bool SmoothServo::isBusy() {
    // return !(this->isRotationOver());
    return this->isBusy_var;
}

bool SmoothServo::isRotationOver() {
    if (this->currentAngle < 0) {
        return false;
    }

    if (this->isCCW == true) {
        return this->currentAngle >= this->destAngle;
    } else {
        return this->currentAngle <= this->destAngle;
    }
}

int16_t SmoothServo::normalizeAngle(uint16_t angle) {
    return max(0, min(180, angle));
}

void SmoothServo::calculateAngularSpeed(uint8_t timeMs) {
    uint32_t tmp32bit = 0;

    /* Check is we still less than at middleAngle */
    if (isMiddleAngleExceded() == false) {
        /* Middle angle still not exceded */
        /* Check if we still need speed up or not */
        if (isMaxSpeedReached() == false) {
            tmp32bit = this->acceleration * timeMs;
            this->currentAngularSpeedMsScale = this->currentAngularSpeedMsScale + tmp32bit;
            
            tmp32bit = (uint32_t)this->minAngularSpeed * MS_FACTOR;
            this->currentAngularSpeedMsScale = max(tmp32bit, this->currentAngularSpeedMsScale);
        }

    } else {
        /* Middle angle already exceded */
        /* Check if we already need to slow speed down */
        if (isSpeedDownAngleExceded()) {
            tmp32bit = this->acceleration * timeMs;
            this->currentAngularSpeedMsScale = this->currentAngularSpeedMsScale - min(this->currentAngularSpeedMsScale, tmp32bit);

            tmp32bit = (uint32_t)this->minAngularSpeed * MS_FACTOR;
            this->currentAngularSpeedMsScale = max(tmp32bit, this->currentAngularSpeedMsScale);
        }
    }
}

int16_t SmoothServo::getAngle(uint8_t timeMs) {
    this->angleChangeTime += timeMs;

    int16_t delta = (uint16_t) ((this->currentAngularSpeedMsScale/MS_FACTOR) * this->angleChangeTime) / MS_FACTOR;

    if (delta > 0) {
        this->angleChangeTime = 0;

        if (isMaxSpeedReached() == false) {
            if (this->isCCW == true) {
                this->speedDownAngle -= delta;
            } else {
                this->speedDownAngle += delta;
            }
        }

        if (this->isCCW == true) {
            return min(this->destAngle, this->currentAngle + delta);
        } else {
            return max(this->destAngle, this->currentAngle - min(this->currentAngle, delta));
        }
    }

    return this->currentAngle;
}

bool SmoothServo::isMiddleAngleExceded() {
    if (this->middleAngle < 0 || this->currentAngle < 0)
        return false;

    if (this->isCCW == true) {
        return this->currentAngle >= this->middleAngle;
    } else {
        return this->currentAngle <= this->middleAngle;
    }
}

bool SmoothServo::isSpeedDownAngleExceded() {
    if (this->speedDownAngle < 0 || this->currentAngle < 0)
        return false;

    if (this->isCCW == true) {
        return this->currentAngle >= this->speedDownAngle;
    } else {
        return this->currentAngle <= this->speedDownAngle;
    }
}

bool SmoothServo::isMaxSpeedReached() {
    return this->currentAngularSpeedMsScale/MS_FACTOR >= this->maxAngularSpeed;
}

/*******************************************************************************
 ****************************     Hardware     *********************************
 ******************************************************************************/

void SmoothServo::setAngle(uint16_t angle) {
    this->currentAngle = angle;
    (this->*write)(angle);
}

void SmoothServo::write_board(uint16_t angle) {
    Board::external()->writeServo(this->servoPin, angle);
}

void SmoothServo::write_arduino(uint16_t angle) {
    this->arduinoServoClient->write(angle);
}

void SmoothServo::write_pca9685(uint16_t angle) {
    /* Convert angle to us */
    uint16_t length = this->servo_min + (float)(this->servo_max - this->servo_min) / 180 * angle;
    length = (uint16_t)((float)length / this->pca_lsb);
    PCA9685::setPWM(this->servoPin, 0, length);
}

void SmoothServo::write_stm8_i2c(uint16_t angle) {
    /* Convert to PWM value */
    STM8_I2C::setPWM(this->servoPin, angle);
}

/*******************************************************************************
 *****************************     Timing     **********************************
 ******************************************************************************/

void SmoothServo::tweakHandler() {
    this->externalTick(this->internalTimerMs);
}

void SmoothServo::externalTick(uint8_t periodMs) {
    /* STEP 1: Check that rotation is still not over */
    if (this->isRotationOver()) {
        this->isBusy_var = false;
        return; 
    }

    /* STEP 2: calculate current angular speed */
    calculateAngularSpeed(periodMs);

    /* STEP 3: calculate new angle */
    int16_t newAngle = getAngle(periodMs);

    if (newAngle != this->currentAngle) {
        setAngle(newAngle);
    }
}