#ifndef SMOOTHSERVO_H
#define SMOOTHSERVO_H

#include <Servo.h>
#include <Tweak.h>
#include "Board.h"

#define SERVO_FREQ  50
#define MS_FACTOR 1000      // 1s = 1000 ms

enum servoEnvironment {
    BOARD,
    ARDUINO_SERVO,
    PCA9685,
    STM8_I2C
};

class SmoothServo : public TweakCallbackOwnerClass {
        /************************** Constructor *******************************/
    public:
        SmoothServo(
            uint8_t pin,
            servoEnvironment mode = BOARD,
            uint16_t min = SERVO_DEFAULT_MIN,
            uint16_t max = SERVO_DEFAULT_MAX,
            uint16_t minSpeed = 30,
            uint16_t maxSpeed = 100,
            uint16_t acceleration = 200
        );
        ~SmoothServo() {}

        /*************************** Hardware *********************************/
    public:
        
    private:
        void setAngle(uint16_t angle);

        void (SmoothServo::* write)(uint16_t angle);
        void write_board(uint16_t angle);
        void write_arduino(uint16_t angle);
        void write_pca9685(uint16_t angle);
        void write_stm8_i2c(uint16_t angle);

        servoEnvironment env;
        Servo* arduinoServoClient = nullptr;
        I2C* i2c_cli = nullptr;
        uint8_t servoPin;
        uint16_t servo_min;
        uint16_t servo_max;
        const float pca_lsb = 4.883;    /* For PCA's 12-bit pwm. 20 ms / 4096 = 4.883 us */


        /*************************** Software *********************************/
    public:
        void begin(uint8_t ms = 0, uint8_t startAngle = 0);
        void setMinAngularSpeed(uint8_t minSpeed);
        void setMaxAngularSpeed(uint16_t maxSpeed);
        void setMaxAcceleration(uint16_t acceleration);
        void rotate(int16_t angle);
        void smoothRotate(int16_t angle);
        bool isBusy();
    private:
        int16_t normalizeAngle(uint16_t angle);
        void calculateAngularSpeed(uint8_t timeMs);
        int16_t getAngle(uint8_t timeMs);
        bool isMiddleAngleExceded();
        bool isSpeedDownAngleExceded();
        bool isMaxSpeedReached();
        bool isRotationOver();

        int16_t destAngle = -1; // Degree
        int16_t currentAngle = -1; // Degree
        int16_t middleAngle = -1; // Degree
        int16_t speedDownAngle = -1; // Degree

        uint16_t minAngularSpeed = 30; // Degree/sec
        uint16_t maxAngularSpeed = 100; // Degree/sec
        uint32_t currentAngularSpeedMsScale = 0; // Degree/msec
        uint16_t acceleration = 200; // Degree/sec^2
        uint16_t angleChangeTime = 0; // Sec

        bool isCCW = false;     // Rotation direction - CW or CCW
        volatile bool isBusy_var = false;

        /**************************** Timing **********************************/
    public:
        void externalTick(uint8_t periodMs);
    private:
        uint8_t internalTimerMs = 0;
        int8_t tweak_id = 0;
        void tweakHandler();
};

#endif /* SMOOTHSERVO_H */